/* AVB support */
/* SPDX-FileCopyrightText: Copyright © 2022 Wim Taymans */
/* SPDX-License-Identifier: MIT */

#ifndef AVB_SRP_H
#define AVB_SRP_H

#include "internal.h"

int avb_srp_register(struct server *server);

#endif /* AVB_SRP_H */
