/* PipeWire */
/* SPDX-FileCopyrightText: Copyright © 2019 Collabora Ltd. */
/*                         @author George Kiagiadakis <george.kiagiadakis@collabora.com> */
/* SPDX-License-Identifier: MIT */

#ifndef PIPEWIRE_EXT_SESSION_MANAGER_H
#define PIPEWIRE_EXT_SESSION_MANAGER_H

/** \defgroup pw_session_manager Session Manager
 * Session manager interface
 */

/**
 * \addtogroup pw_session_manager
 * \{
 */

#include "session-manager/introspect.h"
#include "session-manager/interfaces.h"
#include "session-manager/impl-interfaces.h"
#include "session-manager/keys.h"

/**
 * \}
 */

#endif /* PIPEWIRE_EXT_SESSION_MANAGER_H */
